from __future__ import division
import re

__all__ = ['RedisKeyspaceIterator', 'KeyspaceTracker', 'KeyspaceEmitter']


class RedisKeyspaceIterator(object):
    keyspace_pattern = re.compile(
        r'^([A-Za-z0-9\:\-_]{1,100})\{[A-Za-z0-9_\-\.\+\(\)]{1,500}\}$')
    index_pattern = re.compile(r'^([A-Za-z0-9\:\-_]{1,100})\:[0-9]+\:u\:?$')

    def __init__(self, redis_connection):
        self.conn = redis_connection

    def process(self):
        cursor = 0
        r = self.conn
        patterns = [self.keyspace_pattern, self.index_pattern]
        while True:
            cursor, keys = r.scan(cursor=cursor, count=500)

            for key in keys:
                key = key.decode('utf-8')
                match = None
                for pattern in patterns:
                    match = pattern.match(key)
                    if match:
                        yield match.group(1)
                        break
                if not match:
                    yield '__unknown__'

            if cursor == 0:
                break


class KeyspaceEmitter(object):
    def __init__(self, iterator, callback):
        self.iterator = iterator
        self.callback = callback

    def process(self):
        for keyspace, size in self.iterator.process():
            self.callback(keyspace)
            yield keyspace


class KeyspaceTracker(object):
    def __init__(self, *keyspace_iterators):
        self.keyspaces = {}
        self.total = 0
        self.keyspace_iterators = keyspace_iterators

    def process(self):
        keyspaces = self.keyspaces
        keyspace_iterators = self.keyspace_iterators
        try:
            for iterator in keyspace_iterators:
                for keyspace in iterator.process():
                    try:
                        keyspaces[keyspace] += 1
                    except KeyError:
                        keyspaces[keyspace] = 1
                    self.total += 1
                    yield self.total
        except KeyboardInterrupt:
            pass

    def __enter__(self):
        for _ in self.process():
            pass
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.keyspace_iterators = []

    def stats_output(self):
        total = self.total
        for k, v in sorted(self.keyspaces.items(), key=lambda x: x[1]):
            percentage = (v / total) * 100
            yield "%s: %.1f%%" % (k, percentage)

        yield ""
        yield "sampled: %d" % total



import argparse
import redis
import sys
from signal import signal, SIGTERM
from .api import RedisKeyspaceIterator, KeyspaceTracker

try:
    import redislite
except ImportError:
    redislite = None

__all__ = ['main']


def get_redis_connection(host):
    if redislite and host.startswith("redislite://"):
        return redislite.StrictRedis(host[12:])
    else:
        return redis.StrictRedis.from_url(host)


def process(hosts, out, err=None):
    connections = [get_redis_connection(host) for host in hosts]
    iterators = [RedisKeyspaceIterator(conn) for conn in connections]
    tracker = KeyspaceTracker(*iterators)
    total_count = 0
    for total_count in tracker.process():
        if err and total_count % 1000 == 0:
            err.write(u'\rprocessed: %d' % total_count)

    if err:
        err.write(u'\rprocessed: %d' % total_count)

    out.write(u'\n======================\n')
    for line in tracker.stats_output():
        out.write(u'%s\n' % line)

    out.write(u'======================\n')


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description='iterate through all the keys and collect stats on each '
                    'keyspace')
    parser.add_argument('servers',
                        type=str,
                        nargs='+',
                        help='list of servers in the form of hostname:port')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='turn on verbose output')

    return parser.parse_args(args=args)


# pylint: disable=unused-argument
def sigterm_handler(signum, frame):
    raise SystemExit('--- Caught SIGTERM; Attempting to quit gracefully ---')


def main(args=None, out=None, err=None):
    signal(SIGTERM, sigterm_handler)
    args = parse_args(args=args)
    if err is None:
        err = sys.stderr

    if not args.verbose:
        err = None

    if out is None:
        out = sys.stdout

    process(args.servers, out=out, err=err)
